Lemma plus_0_n : forall n, 0 + n = n.
Proof.
intro n. reflexivity.
Defined.

Lemma plus_n_0 : forall n, n + 0 = n.
Proof.
induction n.
 * reflexivity.
 * simpl. rewrite IHn. reflexivity.
Defined.

(** Ecrire la multiplication pour les entiers *)

Fixpoint mult (n m : nat) :  nat := match n with
| O => 0
| S a => m + mult a m
end.
Notation "n * m" := (mult n m).

(** Prouver les lemmes suivants. *)

Lemma mult_0_n : forall n, 0*n=0.
Proof.
intro n. reflexivity.
Defined.

Lemma mult_n_0 : forall n, n*0= 0.
Proof.
induction n.
  * reflexivity.
  * simpl. assumption.
Defined.

Lemma plus_assoc : forall n m p,
  n+(m+p) = n+m+p.
Proof.
induction n; intros m p.
  * reflexivity.
  * simpl. rewrite IHn. reflexivity.
Defined.

Lemma plus_n_Sm : forall n m,
  n + S m = S (n + m).
Proof.
induction n; intro m.
  * reflexivity.
  * simpl. rewrite IHn. reflexivity.
Defined.

Lemma plus_comm : forall n m, n+m=m+n.
Proof.
induction n; intro m.
  * rewrite plus_0_n. rewrite plus_n_0. reflexivity.
  * simpl. rewrite plus_n_Sm. rewrite IHn. reflexivity.
Defined.

Lemma mult_n_Sm : forall n m, n*S m=n+n*m.
Proof.
induction n; intro m.
  * reflexivity.
  * simpl. rewrite IHn. repeat rewrite plus_assoc.
    rewrite plus_comm with (n:=n) (m:=m). reflexivity.
Defined.

Lemma mult_comm : forall n m, n *m=m*n.
Proof.
induction n; intro m.
  * rewrite mult_0_n. rewrite mult_n_0. reflexivity.
  * simpl. rewrite mult_n_Sm. rewrite IHn. reflexivity.
Defined.

Lemma mult_plus_distr_r : forall n m p,
  (n+m)*p = n*p+m*p.
Proof.
induction n; intros m p.
  * reflexivity.
  * simpl. rewrite IHn. apply plus_assoc.
Defined.

Lemma mult_assoc : forall n m p, n*(m*p) = (n*m)*p.
Proof.
induction n; intros m p.
  * reflexivity.
  * simpl. rewrite mult_plus_distr_r. rewrite IHn.     
    reflexivity.
Defined.

(* Prouver directement *)
Lemma mult_plus_distr_l : forall n m p,
  n*(m+p)=n*m+n*p.
Proof.
induction n as [|n IHn].
+ auto.
+ simpl. intros m p. rewrite IHn. do 2 rewrite <- plus_assoc. 
do 2 rewrite (plus_comm m). do 2 rewrite plus_assoc.
rewrite (plus_comm (n * m) p). auto.
Qed.

(* Prouver en utilisant les lemmes precedents
   (mais bien sur pas [mult_plus_distr_l]) *)
Lemma mult_plus_distr_l_bis : forall n m p,
  n*(m+p)=n*m+n*p.
Proof.
intros n m p.
rewrite mult_comm. rewrite mult_plus_distr_r.
rewrite mult_comm. rewrite plus_comm. 
rewrite mult_comm. rewrite plus_comm. 
reflexivity.
Defined.




Variable A : Type.

Inductive mem : A -> list A -> Prop :=
  mem_head : forall x l, mem x (cons x l)
| mem_tail : forall x y l, 
   mem x l -> mem x (cons y l).

Check app.

Print app.

Hypothesis dec_A : forall (x y : A), {x=y}+{~x=y}.

Lemma app_left : forall l1 l2 x,
mem x l1 -> mem x (app l1 l2).
Proof.
intros l1 l2 x Hmem. induction Hmem.
+ simpl. apply mem_head.
+ simpl. apply mem_tail. assumption.
Qed.

Lemma app_or : forall l1 l2 x, 
mem x (app l1 l2) -> mem x l1 \/ mem x l2.
Proof.
intros l1 l2 x. induction l1 as [|y l1 IHl1].
- simpl. now right.
- simpl. intro Hmem. inversion Hmem as [|x0 y0 l Htail].
  + left. now apply mem_head.
  + destruct (IHl1 Htail).
    * left. now apply mem_tail.
    * now right.
Qed.

(** ** Exercice 3 - Retour un ancien exercice *)

(** Souvenez-vous d'un exercice de la première feuille ! 
    Peut-être que, dans un premier temps, vous aviez défini
    la fonction double ainsi : *)
Definition double_math (m:nat) : nat := plus m m.

(** Mais pour des raisons calculatoires, il était préférable d'écrire : *)
Fixpoint double (m:nat) : nat := 
  match m with
  | 0 => 0
  | S p => S (S (double p))
  end.


(** A présent, vous avez toutes les clefs en main pour démontrer leur
    équivalence : *)

Lemma equiv_double_math_double :
forall n, double_math n = double n.
Proof.
unfold double_math. induction n as [|n IHn].
- auto.
- simpl. rewrite plus_comm. rewrite <- IHn. simpl. reflexivity.
Qed.

(** ** Exercice 4 - Une vision plus calculatoire *)

(** De même que pour [double], il est possible d'écrire une version plus 
    calculatoire de [even] : *)
Definition is_even n := exists p, n = 2 * p.

(** Prouver les lemmes suivants : *)

Lemma zero_is_even : is_even 0.
Proof.
now exists 0.
Qed.


Lemma is_even_plus : forall n p, 
   is_even n -> is_even p -> is_even (n + p).
Proof.
intros n p Hn Hp. destruct Hn as [halfn Hn]. destruct Hp as [halfp Hp].
exists (halfn + halfp). rewrite Hn. rewrite Hp. now rewrite mult_plus_distr_l.
Qed.

(** ** Exercice 5 - Retour sur le prédicat de parité *)

(** On rappelle la définition de [even] sous forme d'un prédicat inductif . *)

Inductive even : nat -> Prop :=
 | even_0 : even 0
 | even_SS : forall n, even n -> even (S (S n)).

(** Prouver les lemmes suivants : *)

Lemma even_4 : even 4.
Proof.
do 2 apply even_SS. apply even_0.
Qed.

Lemma even_succ_lemma : forall n, even n -> even (S (S (S (S n)))).
Proof.
intros. now do 2 apply even_SS.
Qed.

Theorem even_plus4 : forall n, even n -> even (4 + n).
Proof.
intros. simpl. now apply even_succ_lemma.
Qed.

Theorem ev_minus2 : forall n,
  even n -> even (pred (pred n)).
Proof.
intros. destruct n.
- auto.
- destruct n.
  + simpl. apply even_0.
  + now inversion H.
Qed.


Theorem even5_nonsense :
  even 5 -> 2 + 2 = 9.
Proof.
intro H. inversion H as [|n1 H1]. 
inversion H1 as [|n2 H2]. inversion H2.
Qed.

Lemma ev_even : forall n,
  even n -> exists k, n = 2 * k.
Proof.
intros n H. induction H as [|m Hm IHm].
- now exists 0.
- destruct IHm as [k]. exists (S k). rewrite H.
now rewrite mult_n_Sm.
Qed.
 
Lemma even_n_or_Sn : forall n, even n \/ even (S n).
Proof.
induction n as [|n IHn].
- left. constructor.
- destruct IHn.
  + right. now constructor.
  + now left.
Qed.

Lemma even_n_not_Sn : forall n, even n -> ~ even (S n).
Proof.
induction n.
- intros _ H. inversion H.
- intros H H'. apply IHn.
  + now inversion H'.
  + assumption.
Qed.

(** Induction de 2 en 2. *)
(** Faire une induction double sur [P n] revient à faire 
    une induction simple sur ?.
    Pour faire ça, utiliser [assert (H : ?)] puis induction **)

Lemma two_steps_induction : forall P : nat -> Prop, 
   P 0 -> 
   P 1 ->
  (forall n, P n -> P (S (S n))) ->
     forall n, P n.
Proof.
intros P H0 H1 Hn.
assert (forall n : nat, P n /\ P (S n)).
- induction n.
  + split. assumption. assumption.
  + destruct IHn as [IHn1 IHn2]. split.
    * assumption.
    * apply (Hn _ IHn1).
- intro n. now destruct (H n) as [H'].
Qed.

Lemma even_dec : forall n, even n \/ ~ even n.
Proof.
induction n as [|n IHn].
- left. constructor.
- destruct IHn.
  + right. now apply even_n_not_Sn.
  + left. destruct (even_n_or_Sn n).
    * contradiction.
    * assumption.
Qed.

(** Définir [evenb : nat -> bool] qui teste si un entier est pair. *)

Fixpoint evenb (n : nat) : bool := match n with
|0 => true
|1 => false
|S (S n) => evenb n
end.


(** Prouver les lemmes suivants. Si vous avez besoin de résultats sur les
    booléens, vous pouvez les trouver dans [Bool]
    (utiliser [Require Import Bool.]). *)

Lemma evenb_correct : forall n, evenb n = true -> even n.
Proof.
apply (two_steps_induction (fun n : nat => evenb n = true -> even n)).
- constructor.
- simpl. discriminate.
- intros n IHn H. simpl in H. constructor. apply (IHn H).
Qed.

Lemma evenb_complete : forall n, even n -> evenb n = true.
Proof.
apply (two_steps_induction (fun n : nat => even n -> evenb n = true )).
- auto.
- intro H. inversion H.
- simpl. intros n IHn H. apply IHn. now inversion H.
Qed.



Lemma evenb_spec : forall n, evenb n = true <-> even n.
Proof.
intro n. split. apply evenb_correct. apply evenb_complete.
Qed.

(** Prouver [even_dec_bis : forall n, even n \/ ~ even n]
    en utilisant [evenb_spec]. *)

Lemma even_dec_bis : forall n, even n \/ ~ even n.
Proof.
intro n.
assert (H : evenb n = true \/ evenb n = false).
- destruct (evenb n). now left. now right.
- destruct H.
  + left. now apply evenb_correct.
  + right. intro Hn. apply (evenb_complete n) in Hn. rewrite H in Hn. discriminate.
Qed.

(** ** Exercice 6 - Ordre *)

(** [x <= y] est une notation pour [le x y],
    où [le] est défini de manière inductive comme le montre la 
    requête suivante. *)

Print le.

(** Les constructeurs [le_n] et [le_S] qui définissent [le] peuvent être
    être utilisés comme hypothèses / lemmes, par exemple dans la tactique
    [apply]. *)

Goal le 2 3.
Proof.
  apply le_S. apply le_n.
Qed.

(** Si [x <= y] est vrai, il doit avoir été obtenu par l'un des deux 
    constructeurs : la règle d'analyse de cas correspondante est donnée
    par la tactique [inversion]. *)

Goal forall x y : nat, x <= y -> x = y \/ x <= pred y.
Proof.
  intros x y H. inversion H.
  + left. reflexivity.
  + right.
    (* L'étape suivante est facultative,
       elle vous aide à voir ce qui est identique pour Coq :
       [simpl] calcule ce qui peut être calculé. Isa : pas compris*)
    simpl. exact H0.
Qed.

Theorem le_O : forall x:nat, O <= x.
Proof.
induction x.
- constructor.
- constructor. assumption.
Qed.

Lemma le_trans : forall (x y z : nat), 
  x <= y -> y <= z -> x <= z.
Proof.
  intros x y z Hxy Hyz.
  induction Hyz as [|z' Hyz'].
  - assumption.
  - now constructor.
Qed.


Theorem le_S_S : forall x:nat, forall y:nat,
  S x <= S y -> x <= y.
Proof.
intros x y H. inversion H.
- constructor.
- apply (le_trans _ (S x)).
  + do 2 constructor.
  + assumption.
Qed.

(** [x<y] est une notation pour [lt x y] qui est elle-même définie
    en termes de [le]. *)

Print lt.

Theorem le_lt : forall x:nat, forall y:nat, x < S y -> x <= y.
Proof.
unfold lt. apply le_S_S.
Qed.


Theorem lt_S_case : forall x:nat, forall y:nat,
  x < S y -> x = y \/ x < y.
Proof.
unfold lt. intros x y H. inversion H.
- now left.
- now right.
Qed.

(** ** Exercice 7 - Une induction forte *)

(** Avec les ingrédients ci-dessus, vous devriez être capable de prouver
    un principe d'induction forte, exprimé en utilisant une quantification
    sur des prédicats de nombres naturels. *)

Theorem strong_induction :
  forall (P:nat->Prop),
  (forall x:nat, (forall y:nat, y < x -> P y) -> P x) ->
  (forall x:nat, P x).
Proof.
intros P Hstep. 
assert (H' : forall x : nat,
    (forall y : nat, y < x -> P y)).
- induction x.
  + intros y Hy. inversion Hy.
  + unfold lt. intros y Hy. inversion Hy.
    * apply Hstep. apply IHx.
    * now apply IHx.
- intro x. apply Hstep. apply H'.
Qed.

(** ** Exercice 8 - Raisonnement sur les multiples *)

(** Un dernier exercice pour ce tutoriel ! *)

Inductive multiple : nat -> nat -> Prop :=
  | multiple_O : forall n:nat, multiple n O
  | multiple_step :
      forall n:nat, forall m:nat, multiple n m -> multiple n (n+m).
      
      
(* Petit exemple d'à quel point une preuve peut être absconce si on laisse Coq nommer les hypothèses. Je ne suis pas un monstre, j'ai mis des commentaires. *)
Goal forall x:nat,
  multiple 2 x -> multiple 3 x -> multiple 6 x.
Proof.
(* J'utilise l'induction forte, en précisant bien P*)
intro x. apply (strong_induction (fun x => multiple 2 x -> multiple 3 x -> multiple 6 x)).
(*Preuve de l'hérédité*)
intros y H H2 H3.
inversion H3.
  - (*Cas x = 0*) constructor.
  - (* Montrons que m0=y-6 est un multiple de 3 si y est un multiple de 3*)inversion H0.
    + rewrite <- H6 in H4. rewrite <- H4 in H2. inversion H2. inversion H9.
    + (* Montrons que m3=y-6 est un multiple de 2 si y est un multiple de 2*)
      inversion H2. rewrite <- H4 in H9. discriminate. inversion H8. rewrite <- H12 in H10. rewrite <- H10 in H3. inversion H3.
      inversion  H11. rewrite <- H15 in H13. rewrite <- H13 in H10. rewrite <- H10 in H3. inversion H3. inversion H18.
      (*On unifie les variables m0 et m3*)
      assert (Heq : 6+m3 = 6+m0).
      * simpl. simpl in H16. rewrite H16. simpl in H13. rewrite H13. simpl in H10. rewrite H10.
         simpl in H7. rewrite H7. simpl in H4. rewrite H4. reflexivity.
      (* On conclut avec Hstep (ici H) *)
      * rewrite plus_assoc. constructor. simpl. apply H. rewrite <- H4. rewrite <- H7. simpl. repeat constructor.
        simpl in Heq. repeat apply eq_add_S in Heq. rewrite <- Heq. assumption.
        assumption.
Qed.

(** Merci à David Baelde et Catherine Dubois. *)

